/// @addtogroup dynset
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file Rect2Set.hpp
///
/// @author The CAPD Group
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file constitutes a part of the CAPD library, 
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details. 

#ifndef _CAPD_DYNSET_RECT2SET_HPP_ 
#define _CAPD_DYNSET_RECT2SET_HPP_ 

#include <stdexcept>
#include "capd/vectalg/iobject.hpp"
#include "capd/dynset/Rect2Set.h"
#include "capd/matrixAlgorithms/floatMatrixAlgorithms.hpp"

namespace capd{
namespace dynset{

template<typename MatrixType>
Rect2Set<MatrixType>::Rect2Set(const VectorType& x, double sizeFactor /*=20*/) 
  : m_x(x),
    m_r(x.dimension()),
    m_r0(x.dimension()),
    m_B(MatrixType::Identity(x.dimension())),
    m_C(MatrixType::Identity(x.dimension())),
    m_sizeFactor(sizeFactor)
{
  m_r.clear();
  split(m_x,m_r0);
}

template<typename MatrixType>
Rect2Set<MatrixType>::Rect2Set(const VectorType& x,const VectorType& r0, double sizeFactor/*=20*/)
  : m_x(x),
    m_r(x.dimension()),
    m_r0(r0),
    m_B(MatrixType::Identity(x.dimension())),
    m_C(MatrixType::Identity(x.dimension())),
    m_sizeFactor(sizeFactor)  
{
  m_r.clear();
  // we must assure that m_x\in m_x + m_r0
  // m_r is zero, so we verify if m_r is a subset of m_r0
  if(!subset(m_r,m_r0))
  {
    m_x += m_r0;
    split(m_x,m_r0);
  }
}

template<typename MatrixType>
Rect2Set<MatrixType>::Rect2Set(const VectorType& x, const MatrixType& C, const VectorType& r0, double sizeFactor/*=20*/)
  : m_x(x),
    m_r(x.dimension()),
    m_r0(r0),
    m_B(MatrixType::Identity(x.dimension())),
    m_C(C),
    m_sizeFactor(sizeFactor)
{
  m_r.clear();
  // we must assure that m_x\in m_x + m_C*m_r0
  if(!subset(m_r,m_r0))
  {
    VectorType centerR0=m_r0;
    split(centerR0,m_r0);
    // now the m_r0 is centered. We must add C*centerR0 to m_x
    m_x += m_C*centerR0;
    // finally, we split m_x to the point ane the rest is added to r
    split(m_x,m_r);
  }
}

template<typename MatrixType>
Rect2Set<MatrixType>::Rect2Set(
    const VectorType& x,
    const MatrixType& C,
    const VectorType& r0,
    const VectorType& r,
    double sizeFactor/*=20*/
  ):m_x(x),
    m_r(r),
    m_r0(r0),
    m_B(MatrixType::Identity(x.dimension())),
    m_C(C),
    m_sizeFactor(sizeFactor)
{
  VectorType zero(x.dimension());
  if( !( subset(zero,m_r0) && subset(zero,m_r) ) )
  {
    VectorType centerR0=m_r0;
    split(centerR0,m_r0);
    // now the m_r0 is centered. We must add C*centerR0 to m_x
    m_x += m_C*centerR0;
    // finally, we split m_x to the point and the rest is added to r
    m_x += m_r;
    split(m_x,m_r);
  }
}

template<typename MatrixType>
Rect2Set<MatrixType>::Rect2Set(
    const VectorType& x,
    const MatrixType& C,
    const VectorType& r0,
    const MatrixType& B,
    const VectorType& r,
    double sizeFactor/*=20*/
  ):m_x(x),
    m_r(r),
    m_r0(r0),
    m_B(B),
    m_C(C),
    m_sizeFactor(sizeFactor)
{
  VectorType zero(x.dimension());
  if( !( subset(zero,m_r0) && subset(zero,m_r) ) )
  {
    throw std::runtime_error ("Wrong vectors in a constructor of Rect2Set.");
  }
}

template<typename MatrixType>
Rect2Set<MatrixType>::Rect2Set(int dim, double sizeFactor)
  : m_x(dim),
    m_r(dim),
    m_r0(dim),
    m_B(MatrixType::Identity(dim)),
    m_C(MatrixType::Identity(dim)),
    m_sizeFactor(sizeFactor)
{
  m_x.clear();
  m_r.clear();
  m_r0.clear();
}

template<typename MatrixType>
C0Set<MatrixType>* Rect2Set<MatrixType>::clone(void)
{
  return new Rect2Set(*this);
}

template<typename MatrixType>
typename Rect2Set<MatrixType>::ScalarType Rect2Set<MatrixType>::size(void)
{
  return capd::vectalg::size( m_x + m_C*m_r0 + m_B*m_r );
}

template<typename MatrixType>
void Rect2Set<MatrixType>::move(capd::dynsys::DynSys<MatrixType>& dynsys)
{
  move(dynsys,*this);
}

template<typename MatrixType>
void Rect2Set<MatrixType>::move(capd::dynsys::DynSys<MatrixType>& dynsys, Rect2Set<MatrixType>& result) const
{
  MatrixType  S(m_x.dimension(),m_x.dimension());
  
  VectorType xx = m_x + m_C*m_r0 + m_B*m_r;
  
// important: here we assume that both m_r and m_r0 contains zero
// this is assured by each contructor and each step of this algorithm
  result.m_x = dynsys.Phi(m_x) + dynsys.Remainder(xx);
  MatrixType A = dynsys.JacPhi(xx);
// xx is unnecessary now
  capd::vectalg::split(result.m_x,xx);
  result.m_C = A*m_C;
  split(result.m_C,S);
  xx += S*m_r0;
// S is unnecessary now, so we can use it again
  S = A*m_B;
  mid(S,result.m_B); // the same as result.m_B = midMatrix(S), but faster
  capd::matrixAlgorithms::orthonormalize(result.m_B);
  MatrixType Qtr = Transpose(result.m_B);

  result.m_r = (Qtr*S)*m_r + Qtr*xx;
  if(&result!=this)
    result.m_r0 = m_r0;
  
// T.Kapela : we put set r into r0 if r is too big.
  if(m_sizeFactor!=0)
    if( capd::vectalg::size(result.m_r) > m_sizeFactor * capd::vectalg::size(result.m_r0) )
    {
      result.m_r0 = result.m_r +  ((Qtr * result.m_C) * result.m_r0);
      result.m_C = result.m_B;
      result.m_B = MatrixType::Identity(m_x.dimension());
      result.m_r = ScalarType(0.);
    }
}

template<typename MatrixType>
std::string Rect2Set<MatrixType>::show(void) const
{
  std::ostringstream descriptor;
  descriptor << "Rect2Set: x=";
  descriptor << m_x << "\n B=";
  descriptor << m_B << "\n r=";
  descriptor << m_r << "\n C=";
  descriptor << m_C << "\n r0=";
  descriptor << m_r0 << " ";
  return descriptor.str();
}

template<typename MatrixType>
typename Rect2Set<MatrixType>::VectorType Rect2Set<MatrixType>::affineTransformation(
    const MatrixType& A_M, const VectorType& A_C) const
{
  return A_M*(m_x-A_C) + (A_M*m_C)*m_r0 + (A_M*m_B)*m_r;
}

}} // namespace capd::dynset

#endif // _CAPD_DYNSET_RECT2SET_HPP_ 

/// @}
