/// @addtogroup dynset
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file RectSet.hpp
///
/// @author The CAPD Group
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file constitutes a part of the CAPD library, 
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details. 

#ifndef _CAPD_DYNSET_RECTSET_HPP_ 
#define _CAPD_DYNSET_RECTSET_HPP_ 

#include "capd/vectalg/iobject.hpp"
#include "capd/dynset/RectSet.h"
#include "capd/matrixAlgorithms/floatMatrixAlgorithms.hpp"

namespace capd{
namespace dynset{

template<typename MatrixType>
RectSet<MatrixType>::RectSet(int dim) 
  : m_x(dim), 
    m_r(dim), 
    m_B(MatrixType::Identity(dim))
{
  m_x.clear();
  m_r.clear();
}

template<typename MatrixType>
RectSet<MatrixType>::RectSet(const VectorType& x)
  : m_x(x), 
    m_r(x.dimension()), 
    m_B(MatrixType::Identity(x.dimension()))
{
  split(m_x,m_r);
}

template<typename MatrixType>
RectSet<MatrixType>::RectSet(const VectorType& x, const VectorType& r)
  : m_x(x), 
    m_r(r), 
    m_B(MatrixType::Identity(x.dimension()))
{
  if( !subset(VectorType(x.dimension()),m_r) )
  {
    m_x+=m_r;
    split(m_x,m_r);
  }
}

template<typename MatrixType>
RectSet<MatrixType>::RectSet( const VectorType& x, const VectorType& r, const MatrixType& B )
  : m_x(x), m_r(r), m_B(B)
{
  if( !subset(VectorType(x.dimension()),m_r) )
  {
    VectorType centerR = m_r;
    split(centerR,m_r);
    // m_x will be not a single point, but we must assure m_r contains zero
    m_x += m_B*centerR; 
  }
}

template<typename MatrixType>
typename RectSet<MatrixType>::ScalarType RectSet<MatrixType>::size(void)
{
  return capd::vectalg::size(m_x + m_B*m_r);
}

template<typename MatrixType>
void RectSet<MatrixType>::move(capd::dynsys::DynSys<MatrixType>& dynsys)
{
  move(dynsys,*this);
}

template<typename MatrixType>
void RectSet<MatrixType>::move(capd::dynsys::DynSys<MatrixType>& dynsys, RectSet<MatrixType>& result) const
{
  VectorType xx = m_x + m_B*m_r;

  result.m_x = dynsys.Phi(m_x) + dynsys.Remainder(xx);
  MatrixType D = dynsys.JacPhi(xx)*m_B;
// xx is unnecessary, we can use it again  
  split(result.m_x,xx);
  mid(D,result.m_B);
  capd::matrixAlgorithms::orthonormalize(result.m_B);
  MatrixType Qtr = Transpose(result.m_B);
  result.m_r = (Qtr*D)*m_r + Qtr*xx;
}

template<typename MatrixType>
std::string RectSet<MatrixType>::show(void) const
{
  std::ostringstream descriptor;
  descriptor << "RectSet: x=";
  descriptor << m_x << "\n B=";
  descriptor << m_B << "\n r=";
  descriptor << m_r << " ";
  return descriptor.str();
}

template<typename MatrixType>
typename RectSet<MatrixType>::VectorType RectSet<MatrixType>::affineTransformation(
    const MatrixType& A_M, const VectorType& A_C
  ) const
{
  return A_M*(m_x-A_C) + (A_M*m_B)*m_r;
}

}} // namespace capd::dynset

#endif // _CAPD_DYNSET_RECTSET_HPP_ 

/// @}
