/// @addtogroup poincare
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file TimeMapTSC.hpp
///
/// @author Tomasz Kapela
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _CAPD_POINCARE_TIME_MAP_TSC_HPP_
#define _CAPD_POINCARE_TIME_MAP_TSC_HPP_

#include "capd/poincare/TimeMapTSC.h"
#include "capd/interval/DoubleInterval.h"

namespace capd{
namespace poincare{


template <typename DS>
typename TimeMapTSC<DS>::VectorType TimeMapTSC<DS>::operator()(ScalarType time, VectorType v)
// the point on trajectory just after time 'time' - non-rigorous
{
  setNextTimeStep(v);

  ScalarType step = dynamicalSystem.getStep();
  ScalarType nextStep = step;
  ScalarType t(0.0), lastStep;

  while(t + nextStep <= time){
    v = dynamicalSystem(v);
    t+=nextStep;
    nextStep = setNextTimeStep();
  }

  // we make last step
  lastStep = time - t;
  dynamicalSystem.setStep(lastStep);
  v = dynamicalSystem(v);

  dynamicalSystem.setStep(step);
  return v;
}

template<typename DS> template<typename SetType>
typename DS::VectorType TimeMapTSC<DS>::operator()(ScalarType Time, SetType &originalSet)
{
  setNextTimeStep((VectorType)(originalSet));

  ScalarType step = dynamicalSystem.getStep(),
             nextStep = step,
             t(0.0),  lastStep;

  while((t + nextStep) < Time){
    originalSet.move(dynamicalSystem);
    t+=nextStep;
    //nextStep = setNextTimeStep(midVector((VectorType)originalSet));  // it makes small step to produce Taylor expansion
    nextStep = setNextTimeStep();                            // it takes Taylor expansion as it is from previous calculations
  }


  // we make last step
  lastStep = Time - t;
  dynamicalSystem.setStep(lastStep);
  originalSet.move(dynamicalSystem);

  dynamicalSystem.setStep(step);
  return VectorType(originalSet);
}


inline double getPoint(double x)
{
  return x;
}

inline double getPoint(interval x)
{
  return (x.mid().rightBound());
  // return (x.rightBound());
}

template <typename VectorT>
inline double maxNorm(VectorT &v)
{
  double m = getPoint(abs(v[0]));
  for(int i=1; i<v.dimension(); ++i)
    m = capd::max( m,  getPoint(abs(v[i])));
  return m;
}

// this should truncate
inline double truncateTimeStep(double step)
{
  return step;
}


// calculates optimal time step - it changes step of dynamicalSystem
// Do not use this before first step (we need taylor coefficient to be already computed)
template<typename DS>
typename TimeMapTSC<DS>::ScalarType TimeMapTSC<DS>::setNextTimeStep()
{
  int order = dynamicalSystem.getOrder();
  ScalarType optStep(1.e100) , step;
  VectorType *coeff = dynamicalSystem.getCoeff();

  for(int i = order; (i>order - numberOfTerms)&&(i>0); i--)
  {
    //interval coeffNorm = coeff[i].euclNorm();
    ScalarType coeffNorm = maxNorm(coeff[i]);
    step = mid(exp(log(epsilon / coeffNorm)/ i ));
    optStep = capd::min(optStep, step);
  }

  optStep = capd::max(optStep, ScalarType(minTimeStep));
  optStep = truncateTimeStep(getPoint(optStep));
  dynamicalSystem.setStep(optStep);
  return optStep;
}


// Makes small step to calculate Taylor coefficients and then calculates optimal time step
// Usually used before first step.

template<typename DS>
typename DS::ScalarType TimeMapTSC<DS>::setNextTimeStep(VectorType v)
{
  dynamicalSystem.setStep( firstTimeStep );
  dynamicalSystem(v);
  return setNextTimeStep();
}

}} // namespace capd::poincare


#endif // #define _CAPD_POINCARE_TIME_MAP_TSC_HPP_

/// @}
