/// @addtogroup map
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file Node.cpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file constitutes a part of the CAPD library, 
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details. 

#include "capd/interval/DoubleInterval.h"
#include "capd/map/Node.hpp"
#include "capd/multiPrec/MpfrClass.h"
#include "capd/interval/MpInterval.h"

namespace capd{
  namespace map{

template class Node<double>;
template class ConsNode<double>;
template class VarNode<double>;
template class SumNode<double>;
template class DifNode<double>;
template class MulNode<double>;
template class DivNode<double>;
template class SinNode<double>;
template class CosNode<double>;
template class ExpNode<double>;
template class LogNode<double>;
template class SqrtNode<double>;
template class PowNode<double>;

template class MulConsNode<double>;
template class DivConsNode<double>;

template class Node<interval>;
template class ConsNode<interval>;
template class VarNode<interval>;
template class SumNode<interval>;
template class DifNode<interval>;
template class MulNode<interval>;
template class DivNode<interval>;
template class SinNode<interval>;
template class CosNode<interval>;
template class ExpNode<interval>;
template class LogNode<interval>;
template class SqrtNode<interval>;
template class PowNode<interval>;

template class MulConsNode<interval>;
template class DivConsNode<interval>;

template<>
double& PowNode<double>::operator()(int i)
{
  if(i<=m_maxComputedDerivative)
    return value[i];
  m_maxComputedDerivative = i;

  (*left)(i);
  (*right)(i);

  double c=right->value[0];
  if (i==0)
    return value[i] = pow(left->value[i],c);
  if (c==1)
    return value[i] = left->value[i];

  double result=0.;

  if(left->value[0]==0)
  {
    double *temp1 = new double[i+1];
    double *temp2 = new double[i+1];

    int j,k,p;
    for(j=0;j<=i;++j)
      temp1[j]=left->value[j];

    for(k=1;k<c;++k)
    {
      for(p=0;p<=i;++p)
      {
        temp2[p]=0.;
        for(j=0;j<=p;++j)
          temp2[p] += left->value[j]*temp1[p-j];
      }
      for(p=0;p<=i;++p)
        temp1[p]=temp2[p];
      }

      result = temp1[i];
      delete []temp1;
      delete []temp2;
      return value[i] = result;
    }

  for(int j=0;j<i;++j)
    result+= (i*c-j*(c+1)) * value[j] * left->value[i-j];

  return value[i] = result / (i * left->value[0]);
}

template<>
long double& PowNode<long double>::operator()(int i)
{
  if(i<=m_maxComputedDerivative)
    return value[i];
  m_maxComputedDerivative = i;

  (*left)(i);
  (*right)(i);

  long double c=right->value[0];
  if (i==0)
    return value[i] = pow(left->value[i],c);
  if (c==1)
    return value[i] = left->value[i];

  long double result=0.;

  if(left->value[0]==0)
  {
    long double *temp1 = new long double[i+1];
    long double *temp2 = new long double[i+1];

    int j,k,p;
    for(j=0;j<=i;++j)
      temp1[j]=left->value[j];

    for(k=1;k<c;++k)
    {
      for(p=0;p<=i;++p)
      {
        temp2[p]=0.;
        for(j=0;j<=p;++j)
          temp2[p] += left->value[j]*temp1[p-j];
      }
      for(p=0;p<=i;++p)
        temp1[p]=temp2[p];
      }

      result = temp1[i];
      delete []temp1;
      delete []temp2;
      return value[i] = result;
    }

  for(int j=0;j<i;++j)
    result+= (i*c-j*(c+1)) * value[j] * left->value[i-j];

  return value[i] = result / (i * left->value[0]);
}

#ifdef __HAVE_MPFR__

typedef capd::multiPrec::MpfrClass Real;
typedef capd::intervals::Interval<Real> MpInterval;

template class Node<Real>;
template class ConsNode<Real>;
template class VarNode<Real>;
template class SumNode<Real>;
template class DifNode<Real>;
template class MulNode<Real>;
template class DivNode<Real>;
template class SinNode<Real>;
template class CosNode<Real>;
template class ExpNode<Real>;
template class LogNode<Real>;
template class SqrtNode<Real>;
template class PowNode<Real>;

template class MulConsNode<Real>;
template class DivConsNode<Real>;

template class Node<MpInterval>;
template class ConsNode<MpInterval>;
template class VarNode<MpInterval>;
template class SumNode<MpInterval>;
template class DifNode<MpInterval>;
template class MulNode<MpInterval>;
template class DivNode<MpInterval>;
template class SinNode<MpInterval>;
template class CosNode<MpInterval>;
template class ExpNode<MpInterval>;
template class LogNode<MpInterval>;
template class SqrtNode<MpInterval>;
template class PowNode<MpInterval>;

template class MulConsNode<MpInterval>;
template class DivConsNode<MpInterval>;

template<>
Real& PowNode<Real>::operator()(int i)
{
  if(i<=m_maxComputedDerivative)
    return value[i];
  m_maxComputedDerivative = i;

  (*left)(i);
  (*right)(i);

  Real c=right->value[0];
  if (i==0)
    return value[i] = (left->value[i]).pow(c);
  if (c==1)
    return value[i] = left->value[i];

  Real result=0.;

  if(left->value[0]==0)
  {
    Real *temp1 = new Real[i+1];
    Real *temp2 = new Real[i+1];

    int j,k,p;
    for(j=0;j<=i;++j)
      temp1[j]=left->value[j];

    for(k=1;k<c;++k)
    {
      for(p=0;p<=i;++p)
      {
        temp2[p]=0.;
        for(j=0;j<=p;++j)
          temp2[p] += left->value[j]*temp1[p-j];
      }
      for(p=0;p<=i;++p)
        temp1[p]=temp2[p];
      }

      result = temp1[i];
      delete []temp1;
      delete []temp2;
      return value[i] = result;
    }

  for(int j=0;j<i;++j)
    result+= (i*c-j*(c+1)) * value[j] * left->value[i-j];

  return value[i] = result / (i * left->value[0]);
}

#endif

  }} // namespace capd::map
/// @}
